use crate::model::{ECCKey, Peer, PeerBuilder, WireguardController};
use std::io::{BufRead, BufReader, Error, ErrorKind, Result, Write};
use std::net::{IpAddr, SocketAddr};
use std::os::unix::net::UnixStream;
use std::path::PathBuf;
use std::time::Duration;

pub struct Userspace(PathBuf);

impl Userspace {
    pub fn new<P: Into<PathBuf>>(path: P) -> Userspace {
        Userspace(path.into())
    }
}

impl WireguardController for Userspace {
    fn peers<'a>(&'a mut self) -> Result<Box<Iterator<Item = Result<Peer>> + 'a>> {
        let mut stream = UnixStream::connect(&self.0)?;
        stream.write_all(b"get=1\n")?;

        fn build_peer(builder: &mut PeerBuilder, line: Result<String>) -> Option<Result<Peer>> {
            let line = match line {
                Ok(line) => line,
                Err(e) => return Some(Err(e)),
            };
            fn parse_err<O, E>(res: std::error::Result<O, E>) -> Result<O> {
                res.map_err(|err: String| Error::new(ErrorKind::InvalidData, err.into()))
            }
            let mut iter = line.chars();
            let key = iter.by_ref().take_while(|c| c != &'=').collect::<String>();
            let value = iter.collect::<String>();
            let value_as_num = || parse_err(value.parse::<u64>());
            let mut peer: Option<Peer> = None;
            let add_key = |key| {
                if builder.is_whole() {
                    peer = Some(parse_err(builder.build())?);
                    *builder = PeerBuilder::default();
                } else {
                    peer = None
                }
                builder.key(key);
                Ok(peer)
            };
            match key.as_ref() {
                "public_key" => {
                    add_key(parse_err(ECCKey::from_base64(value))?)?;
                }
                /*"private_key" => {
                    add_key(ECCKey::from_base64(value)?)?;
                }
                "endpoint" => {
                    builder.endpoint(Some(parse_err(value.parse::<SocketAddr>())?));
                }
                "last_handshake_time_sec" => {
                    builder.add_last_handshake(Duration::from_secs(value_as_num()?));
                }
                "last_handshake_time_nsec" => {
                    builder.add_last_handshake(Duration::from_nanos(value_as_num()?.into()));
                },
                "persistent_keepalive" => {
                    builder.persistent_keepalive(Some(Duration::from_secs(value_as_num()?.into())));
                },*/
                "allowed_ip" => {
                    let mut parts = value.split("/").into_iter();
                    let net = match (
                        parts.next().and_then(|addr| addr.parse::<IpAddr>().ok()),
                        parts.next().and_then(|mask| mask.parse::<u8>().ok()),
                    ) {
                        (Some(addr), Some(mask)) => Some((addr, mask)),
                        (Some(addr), None) if addr.is_ipv6() => Some((addr, 128)),
                        (Some(addr), None) => Some((addr, 32)),
                        _ => None,
                    };
                    if let Some(net) = net {
                        builder.add_allowed_ip(net);
                    }
                }
            }
            peer.map(|peer| Ok(peer))
        }

        let peers = BufReader::new(stream)
            .lines()
            .scan(PeerBuilder::default(), build_peer);
        Ok(Box::new(peers))
    }

    fn update_peer(&mut self, peer: &Peer) -> Result<()> {
        loop {}
    }
}
